/* -*-  Mode:C; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 1999, 2000 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `LICENSE' in the Linux source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file CMAUTHORS which should have accompanied this software 
 * for a list of all copyright holders. 
 * 
 * Particular files in this distribution may be derived from previously
 * copyrighted software. This copyright applies only to those changes made by
 * the copyright holders listed in the CMAUTHORS file.
 */
/*
 * An implementation of the Congestion Manager (CM) for Linux.
 * This file implements the CM congestion controller modeled after 
 * TCP.  See RFC 2581 for details on the specific mechanisms.
 */


#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <net/cm.h>
#include <net/cm_tcpcongctl.h>
#include <linux/wanpipe.h>
#include <net/cm_debug.h>

/*
 * Initialize congestion controller state when a new CM is instantiated 
 * for this destination (macroflow).  Called by cm_open().
 */
void tcpcc_init(cm_congctl_t *tcpcc)
{
        CM_DPRINTF(CM_DEBUG_FUNCALL|CM_DEBUG_MACROFLOW_INIT,
		   "cm_debug:  %s:  initting CC %p for a new macroflow\n",
		   __FUNCTION__, tcpcc);

	tcpcc->state.size = tcpcc->mflow->mtu;
	/* 
	 * Nothing sent so far to set things to 0 or -1 (in some cases).
	 */
	tcpcc->state.outstanding = 0;	
	tcpcc->state.last_size = 0;
	tcpcc->state.last_time = 0;
	tcpcc->state.last_loss = 0;
	tcpcc->state.srtt = -1;
	tcpcc->state.rttdev = -1;
	tcpcc->state.rttmin = 0;
	tcpcc->state.rate = 0;
	tcpcc->state.cwnd = tcpcc->mflow->mtu;
	tcpcc->state.ssthresh = CM_SSTHRESH; /* initial slow-start thresh */

	tcpcc->ops.update = tcpcc_update;
	tcpcc->ops.query = tcpcc_query;
	tcpcc->ops.notify = tcpcc_notify;
	tcpcc->ops.wakeup = tcpcc_wakeup;
	printk("cm_tcpcongctl.c: initialization done\n");
}

/*
 * Called in first cm_update() call.
 */
inline void tcpcc_startup(cm_congctl_t *tcpcc, int nrecd, double rtt, 
			    int lossmode)
{
        CM_DEBUG_ENTER();
	if (rtt > 0) {
		tcpcc->state.srtt = tcpcc->state.rttmin = rtt;
		tcpcc->start.rttdev = 0; /* first sample ==> no deviation! */
	}
}

/* 
 * Go back into slow start, e.g., when CM_PERSISTENT is reported.
 */
inline void tcpcc_restart(cm_congctl_t *tcpcc, u32 now)
{
	tcpcc->state.cwnd = tcpcc->state.size; /* 1 packet */
	tcpcc->state.ssthresh = tcpcc->state.cwnd;
	tcpcc->state.last_loss = now;
	CM_DPRINTF(CM_DEBUG_CONGCTL,
		   "TCP_RESTART %d cwnd %d ssthresh %d \n",
		   now, tcpcc->state.cwnd, tcpcc->state.ssthresh);
	cm_log(CMLOG_RESTART,now,tcpcc->state.cwnd,tcpcc->state.ssthresh);
}

/*
 * Multiplicative (factor-of-two) reduction in congestion window.
 * Happens at most once per RTT period, to prevent multiple reductions
 * in the same congestion "epoch".
 */
inline void tcpcc_decrease(cm_congctl_t *tcpcc, u32 now)
{
	if (now - tcpcc->state.last_loss < tcpcc->state.srtt)
		return;		/* already taken care of reduction */

	tcpcc->state.cwnd=(tcpcc->state.outstanding>>CM_DECR_SHIFT);
	if (tcpcc->state.cwnd < tcpcc->state.size)
		tcpcc->state.cwnd = tcpcc->state.size;
	tcpcc->state.ssthresh = tcpcc->state.cwnd;
	tcpcc->state.last_loss = now;
	CM_DPRINTF(CM_DEBUG_CONGCTL, "TCP_DECREASE %d cwnd %d ssthresh %d\n",
		   now, tcpcc->state.cwnd, tcpcc->state.ssthresh);
	cm_log(CMLOG_DECREASE, now, tcpcc->state.cwnd, tcpcc->state.ssthresh);
}

/*
 * TCP-style additive-increase.  Some important differences from RFC 2581:
 * (i)  Byte-counting to account for infrequent feedback.
 * (ii) Bursts are avoided by ensuring that sender paces (shapes) packets out.
 *      nrecd is number of successfully received bytes.
 */
inline void tcpcc_increase(cm_congctl_t *tcpcc, int nrecd)
{
	__u32 now = (__u32)(jiffies);
	if (tcpcc->state.cwnd >= tcpcc->state.ssthresh)
		tcpcc->state.cwnd += 
			min ((int)((nrecd*tcpcc->state.size)/
				   tcpcc->state.cwnd), tcpcc->state.size); 
	else			/* don't jump over ssthresh */
		tcpcc->state.cwnd = min(tcpcc->state.cwnd + nrecd, 
					tcpcc->state.ssthresh);
	CM_DPRINTF(CM_DEBUG_CONGCTL, "TCP_INCREASE %d cwnd %d ssthresh %d\n",
		   now,tcpcc->state.cwnd, tcpcc->state.ssthresh);
	cm_log(CMLOG_INCREASE, now, tcpcc->state.cwnd, tcpcc->state.ssthresh);
}

/* 
 * Estimate round-trip time.
 */
inline void tcpcc_update_rtt(cm_congctl_t *tcpcc, u32 rtt)
{
	u32 shrinkage;
	
	if (rtt == 0)
		return;
	/* dividing by 8 and multiplying by 7 tends to result in 0 too often */
	/* this is equivalent and subtraction isn't too expensive */
	shrinkage = tcpcc->state.srtt >> CM_RTTSHIFT;
	tcpcc->state.srtt = tcpcc->state.srtt - shrinkage;
	tcpcc->state.srtt += rtt >> CM_RTTSHIFT;
	if (rtt < tcpcc->state.rttmin)
		tcpcc->state.rttmin = rtt;

	/* XXX: need to hack in rttdev calculation too */
}

void tcpcc_update(cm_congctl_t *tcpcc, int nsent, int nrecd, 
		 u32 rtt, int lossmode)
{
	__u32 now = (__u32)(jiffies);
	int changed = 0;
	int outs = tcpcc->state.outstanding;
	int sent;

	if (nsent == 0) 
		return;
	CM_DPRINTF(CM_DEBUG_CONGCTL_VERBOSE,
		   "update: outstanding changes from %d to %d \n",
		   tcpcc->state.outstanding,
		   tcpcc->state.outstanding-nsent);
	tcpcc->state.outstanding -= nsent; /* um... what about reordering? */

	if (tcpcc->state.srtt == 0 && rtt != 0) /*&& lossmode == CM_NOLOSS)*/
		/* XXX */
		tcpcc_startup(tcpcc, nrecd, rtt, lossmode);/* slow start */

	/* Assuming no reordering fixes.  Need to modify for reordering. */
	switch (lossmode) {
	case (CM_ECN):
	case (CM_TRANSIENT):
		tcpcc_decrease(tcpcc, now);
		changed = 1;
		tcpcc->state.last_backoff = 0;
		tcpcc->state.backoff = 1;
		break;
	case (CM_PERSISTENT):
		tcpcc_restart(tcpcc, now);
		changed = 1;
		tcpcc->state.last_backoff = 0;
		tcpcc->state.backoff = 1;
		break;
/*
	case (CM_PARTIALACK):
		Don't do anything for partial acks.
		break;
*/
	default:  /* no losses */
		tcpcc_update_rtt(tcpcc, rtt);
		/* Commented Deepak: dont know why? */
		/* In last test, commented because wasn't allowing
                 * inc. in many cases */
		/*
		if ((tcpcc->state.outstanding + nsent) > 
		    ((tcpcc->state.cwnd /2))){
		*/
			tcpcc_increase(tcpcc, nrecd);
		/*
		}
		*/
		changed = 1;
		tcpcc->state.last_backoff = 0;
		tcpcc->state.backoff = 1;
		break;
	}

	tcpcc_update_rate(tcpcc);

	/* schedule different flow on mflow*/
	cmcc_wakeup(tcpcc);
	sent = tcpcc->state.outstanding - (outs-nsent);
	CM_DPRINTF(CM_DEBUG_CONGCTL_VERBOSE,
		   "acked %d sent %d\n",  nsent, sent);
/*	if (sent == 0) {
		int pacing_est = tcpcc->state.size;
		if (tcpcc->state.rttmin != 0) {
			pacing_est = (tcpcc->state.cwnd * 
				      (now - tcpcc->state.last_time + 1)) / 
				tcpcc->state.rttmin;
		}
		printk("pacing = %d last %d cwnd %d outs %d\n",
		       pacing_est, tcpcc->state.last_size,
		       tcpcc->state.cwnd, tcpcc->state.outstanding);
	}
*/		
	/* cms_schedule(tcpcc->mflow->sched); */

	/*	if (changed)
		cms_update_stats(tcpcc->mflow->sched); */
}

/*
 * Fill in the bandwidth, srtt and rtt deviation for the macroflow.
 */
void tcpcc_query(cm_congctl_t *tcpcc, double *bw, double *srtt,
		   double *rttdev)
{
	*bw = tcpcc->state.rate;
	*srtt = tcpcc->state.srtt;
	*rttdev = tcpcc->state.rttdev;
}

void tcpcc_notify(cm_congctl_t *tcpcc, int nsent)
{
	__u32 now = (__u32)(jiffies);

	CM_DPRINTF(CM_DEBUG_CONGCTL_VERBOSE,
		   "notify: outstanding changes from %d to %d \n",
		   tcpcc->state.outstanding,
		   tcpcc->state.outstanding+nsent);
	tcpcc->state.outstanding += nsent;
	if ( (tcpcc->state.rttmin == 0) ||
	     (((now - tcpcc->state.last_time) 
	      * tcpcc->state.cwnd / tcpcc->state.rttmin)
	     > tcpcc->state.last_size)) {
		/* pace is slower than required */
		tcpcc->state.last_time = now;
		tcpcc->state.last_size = nsent;
	} else {
		/* sent too soon... need to compensate on next send */
		tcpcc->state.last_size += nsent;/* so inflate last_size */
	}
}

void tcpcc_wakeup(cm_congctl_t *tcpcc) 
{
	cm_sched_t *sched;
#if 0
	__u32 now = (__u32)(jiffies);
#endif
	int pacing_est;
	int maxburst;
	int outstanding;
	sched = tcpcc->mflow->sched;
	pacing_est = tcpcc->state.size;
	maxburst = CM_TCPCC_MAXBURST;
	outstanding = tcpcc->state.outstanding;

#if 0
	if (tcpcc->state.rttmin != 0) {
		pacing_est = (tcpcc->state.cwnd * 
			      (now - tcpcc->state.last_time + 1)) / 
			tcpcc->state.rttmin;
		if (pacing_est < tcpcc->state.last_size)
			/* too soon for pacing */
			return;
		pacing_est -= tcpcc->state.last_size;
	} 
#endif
	while (/* (pacing_est >= tcpcc->state.size) && */
	       ((tcpcc->state.cwnd >= outstanding + tcpcc->state.size)) &&
	       (maxburst-- > 0)) {
		/* send a pkt */
		cms_schedule(sched);
		pacing_est -= tcpcc->state.size;
		outstanding += tcpcc->state.size;
	}

}

/* 
 * Compute rate on mflow --- either cwnd or srtt has changed.
 * Then check each flow on the mflow:
 * - if rate callback bounded by thresholds, check them and do callback;
 * - if there aren't any thresholds, do the callback anyway.
 */
void tcpcc_update_rate(cm_congctl_t *tcpcc)
{
        cm_macroflow_t *mflow = tcpcc->mflow;
	int nonzero_srtt = tcpcc->state.srtt;
	cm_flow_t *flow;
	double newrate, newfrate;
	int irate;
	if (tcpcc->state.srtt == 0) nonzero_srtt = CM_RTT_DEFAULT;
#ifdef CM_AGING
	tcpcc->state.rate = (double) tcpcc->state.cwnd/(nonzero_srtt * 
					    tcpcc->state.backoff);
#else
	tcpcc->state.rate = (double) tcpcc->state.cwnd/nonzero_srtt;
#endif
	irate = tcpcc->state.rate;
	CM_DPRINTF(CM_DEBUG_CONGCTL,
		   "mflow rate = %d cwnd %d srtt %d\n",
		   irate, tcpcc->state.cwnd, nonzero_srtt);
	newrate = tcpcc->state.rate;
	flow = mflow->flows;
	while (flow) {
		if (!(flow->downthreshmult < 0.0)) { /* if enabled */
		   newfrate = newrate * cms_query_share(mflow->sched,flow);
		   if (newfrate < flow->downthreshmult*flow->lastrate ||
	    	    	newfrate > flow->upthreshmult*flow->lastrate) {
			flow->lastrate = newfrate;
			flow->ratechange = 1;
			if (flow->do_callback) cmapp_send_wakeup(flow);
		   }
		}
		flow = flow->next;
	}
}


int cm_log_data[CM_LOG_SIZE];
int cm_log_index = 0;

void cm_log(int event, u32 time, int arg1, int arg2)
{
	cm_log_data[cm_log_index] = event;
	cm_log_data[cm_log_index+1] = time;
	cm_log_data[cm_log_index+2] = arg1;
	cm_log_data[cm_log_index+3] = arg2;
	cm_log_index += 4;
#if 0
	printk("index %d event %d time %d arg1 %d arg2 %d\n", cm_log_index, event, time, arg1, arg2);
#endif
	/* assumes size of log is multiple of number of entries per call... */
	if (cm_log_index >= CM_LOG_SIZE) cm_log_index = 0;
}
